import logging
import os
import string
import random
import subprocess
from OpenSSL import crypto
from pathlib import Path
from werkzeug.utils import secure_filename

from .uci import setUciConfigs
from .exceptions import InvalidUploadedFileError
from .constants import *

log = logging.getLogger('werkzeug')

def generateSSLCerts():
    # if certs dir does not exist, create it
    p = Path(SSL_CERTIFICATES_DIR)
    if not p.exists():
        try:
            p.mkdir()
        except Exception as e:
            log.error(e)
            raise e

    # if certificate do not exist, generate it
    if os.path.isfile(os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem')):
        log.debug("SSL certificate already exists, do not need to create")
    else:
        try:
            # generate RSA key pair
            k = crypto.PKey()
            k.generate_key(crypto.TYPE_RSA, 2048)

            emailAddress = "info@barix.com"
            countryName = "CH"
            stateOrProvinceName = "Zurich"
            localityName = "Dubendorf"
            organizationName = "Barix AG"
            commonName = "barix.local"
            cert = crypto.X509()
            cert.get_subject().C = countryName
            cert.get_subject().ST = stateOrProvinceName
            cert.get_subject().L = localityName
            cert.get_subject().O = organizationName
            cert.get_subject().CN = commonName
            cert.get_subject().emailAddress = emailAddress
            #cert.set_serial_number(serialNumber)
            cert.gmtime_adj_notBefore(0)
            cert.gmtime_adj_notAfter(20*365*24*60*60)
            cert.set_issuer(cert.get_subject())
            cert.set_pubkey(k)
            cert.sign(k, 'sha256')

            with open(os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem'), "w") as f:
                f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k).decode("utf-8"))

            with open(os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem'), "a") as f:
                f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"))

            # set chmod of certificate file
            os.chmod(os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem'), 0o400)

            # set uci value
            setUciConfigs({'httpd.ssl.certificate':os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem')}, restartServices=False)
        except Exception as e:
            log.error("Error while generating SSL certificates: {}".format(e))
            raise e

def getCACertificates():
    try:
        certificatesList = []
        process = subprocess.run(['custom-ca-mgr', 'status'], capture_output=True)
        output = process.stdout
        decodedOutput = output.decode("utf-8")
        lines = decodedOutput.split('\n')
        for line in lines:
            elems = line.split(' ')
            if len(elems) > 1: # last \n counts as one entire line
                #print(elems)
                filename = elems[1]
                #print(CA_CERTIFICATES_FOLDER)
                filePath = os.path.join(CA_CERTIFICATES_DIR, filename)
                #print(filePath)
                fileSize = os.path.getsize(filePath)
                state = elems[0]
                certificatesList.append({"name": filename, "size": fileSize, "state": state})
    except Exception as e:
        log.error(e)
        raise e
    else:
        log.debug("Getting CA Certificates: {}".format(certificatesList))
        return certificatesList

def uploadCACertificate(filesUploaded):
    try:
        if 'certificate' not in filesUploaded:
            raise InvalidUploadedFileError('Invalid file')
        uploadedFile = filesUploaded['certificate']
        # if user does not select file, browser also
        # submit an empty part without filename
        if uploadedFile.filename == '':
            raise InvalidUploadedFileError('Invalid filename')
        # check file size
        # generate random string
        letters = string.ascii_lowercase
        tmp_filename = ''.join(random.choice(letters) for i in range(8))
        tmp_path = os.path.join('/tmp', tmp_filename)

        uploadedFile.save(tmp_path)
        filesize = os.stat(tmp_path).st_size
        if filesize < 133 or filesize > 16384000:
            raise InvalidUploadedFileError("File size is invalid. Please make sure you're uploading the right file.")
        # check file extension
        ALLOWED_EXTENSIONS = {'crt'}
        if '.' in uploadedFile.filename and uploadedFile.filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS:
            try:
                addCACertificates(tmp_path, secure_filename(uploadedFile.filename))
                return uploadedFile.filename + " installed."
            except OSError:
                raise InvalidUploadedFileError('Invalid file.')
            except Exception as e:
                raise e
        else:
            raise InvalidUploadedFileError('Format not supported.')
    except Exception as e:
        raise e

def addCACertificates(tmp_filepath, filename):
    try:
        #print('openssl x509 -in /tmp/foo')
        process = subprocess.run(['openssl', 'x509', '-in', tmp_filepath], capture_output=True)
        stderr = process.stderr
        empty = ''.encode("utf-8")
        if stderr != empty:
            raise OSError
    except Exception as e:
        log.error(e)
        raise e
    try:
        final_filepath = os.path.join(CA_CERTIFICATES_DIR, filename)
        subprocess.run(['mv', tmp_filepath, final_filepath])
        #print('custom-ca-mgr', 'refresh')
        log.debug("Certificate {} installed successfully.".format(filename))
        subprocess.run(['custom-ca-mgr', 'refresh'])
    except Exception as e:
        log.error(e)
        raise e

def removeCACertificates(certificatesList):
    try:
        counter = 0
        for certificateName in certificatesList:
            filePath = os.path.join(CA_CERTIFICATES_DIR, certificateName)
            if os.path.isfile(filePath):
                os.remove(filePath)
                log.debug("Certificate {} removed".format(filePath))
                counter += 1
        subprocess.run(['custom-ca-mgr', 'refresh'], capture_output=False)
    except Exception as e:
        log.error(e)
        raise e
    else:
        return counter